/*
 * NAME:
 *	desenc - routines to interface to a DES library
 *
 * SYNOPSIS:
 *	void    de_setkeys(u_char *rkey, u_char *wkey)
 *	u_char *de_buf_ksched(des_key_schedule ksched, \\
 *		des_cblock *iv, u_char *in, u_char *out, \\
 *		int nbytes, int encrypt)
 *	u_char *de_buf_key(u_char *key, u_char *in, \\
 *		u_char *out, int nbytes, int encrypt)
 *	u_char *de_buf(u_char *buf, int nbytes, int encrypt)
 *	u_char *de_encrypt(u_char *buf, int nbytes)
 *	u_char *de_decrypt(u_char *buf, int nbytes)
 *	int     de_status(char **mp)
 *	void    de_setstatus(int s)
 *	void    de_on(int fd)
 *	void    de_off(void)
 *	
 * DESCRIPTION:
 *	This module provides a simple interface to a DES library (Eric
 *	Young\'s in this case).  Only 'de_buf_ksched' and 'de_setkeys'
 *	need be changed to support a different library.
 *
 * BUGS:
 *	Couldn\'t think of a better prefix than 'de_'.
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */

#ifndef lint
static char RCSid[] = "$Id: desenc.c,v 1.5 1997/05/15 15:51:38 sjg Exp $";
#endif
#include <stdio.h>
#include <sys/types.h>

#include "des.h"

#include "desenc.h"

#include "sslfd.h"

static des_key_schedule RKsched, WKsched;
static des_cblock Rvec, Wvec;

#if defined(MAIN) && !defined(DEBUG)
# define DEBUG
#endif

static int DeStatus = DE_OK;
static int DeDo = 0;
static int De2keys = 0;

/* safer to leave this in */
FILE *DeDumpLog = NULL;

#ifdef DEBUG
# define dump_buf(hdr, p, n)	if (DeDumpLog) { fprintf(DeDumpLog, "%s:", hdr); fhex_dump(DeDumpLog, p, n); }
#else
# define dump_buf(hdr, p, n)
#endif

#define CLEAR(a)	memset(a, 0, sizeof (a));

/*
 * This just lets us re-initialize the engine,
 * if called with NULL arg, we re-load stored key.
 */

void
de_setkeys(rkey, wkey)
	u_char *rkey, *wkey;
{
	static int once = 1;

	if (once && rkey && wkey) {
#ifndef AUTH_OLDHPW
		once = 0;
#endif
		CLEAR(Rvec);
		CLEAR(Wvec);
	
		DeStatus |= DE_KEYSET;

		De2keys = (rkey != wkey);
		
		des_set_key((des_cblock *)rkey, RKsched);
		des_set_key((des_cblock *)wkey, WKsched);
	}
}


u_char *
de_buf_ksched(ksched, iv, in, out, nbytes, encrypt)
	des_key_schedule ksched;
	des_cblock *iv;
	u_char *in, *out;
	int nbytes, encrypt;
{
#define HUNK 512
	u_char temp[HUNK + 1];
	int i, n;
	
	for (i = 0; i < nbytes; i += n) {
		if ((n = nbytes - i) > HUNK)
			n = HUNK;
		/*
		 * des_cfb_encrypt() can handle up to 64 bits at a
		 * time, but 8 is what you _need_ to handle odd sized
		 * buffers safely...
		 */
		des_cfb_encrypt(&in[i], temp, 8, n,
				ksched, iv, encrypt);
		memcpy((char *)&out[i], temp, n);
	}
	return out;
}

/*
 * set a key and encrypt/decrypt
 */
u_char *
de_buf_key(key, in, out, nbytes, encrypt)
	u_char *key, *in, *out;
	int nbytes, encrypt;
{
	des_key_schedule ksched;
	des_cblock iv;
	
	des_set_key((des_cblock *)key, ksched);
	CLEAR(iv);
	
	return de_buf_ksched(ksched, &iv, in, out, nbytes, encrypt);
}

/*
 * Simply do an in-place cfb mode {de,en}cryption on buf.
 * Must have called de_setkeys() by now...
 */
u_char *
de_buf(buf, nbytes, encrypt)
	u_char *buf;
	int nbytes, encrypt;
{
	return de_buf_ksched((encrypt) ? WKsched : RKsched,
			     (encrypt) ? &Wvec : &Rvec,
			     buf, buf, nbytes, encrypt);
}


int
de_encrypt(fd, buf, nbytes)
	int fd;
	u_char *buf;
	int nbytes;
{
	if (DeDo) {
		dump_buf("Encrypt", buf, nbytes);

		de_buf(buf, nbytes, 1);
		
		dump_buf("Result", buf, nbytes);
	}
	return nbytes;
}


int
de_decrypt(fd, buf, nbytes)
	int fd;
	u_char *buf;
	int nbytes;
{
	if (DeDo) {
		dump_buf("Decrypt", buf, nbytes);
		
		de_buf(buf, nbytes, 0);
		
		dump_buf("Result", buf, nbytes);
	}
	return nbytes;
}

int
de_status(mp)
	char **mp;
{
	static char m[32];
	char *s = "DES-CFB";

	if (mp) {
		sprintf(m, "\r\n%s%s Encryption is %s.\r\n",
			s,
			De2keys ? "x2" : "",
			DeDo ? "on" : "off");
		*mp = m;
	}
	
	return DeStatus;
}


void
de_setstatus(s)
	int s;
{
	DeStatus = s;
}

void
de_on(fd)
	int fd;
{
	if (DeStatus & DE_KEYSET) {
		DeDo = 1;
	
		set_hook(fd, HK_READ, de_decrypt);
		set_hook(fd, HK_WRITE, de_encrypt);
		DeStatus |= DE_CRYPTING;
	}
}


void
de_off(fd)
	int fd;
{
	DeDo = 0;

	set_hook(fd, 0, 0);			/* reset */
	DeStatus &= ~DE_CRYPTING;
}

#ifdef MAIN
int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int optind, getopt();
	extern char *optarg, *basename();
	int i;
	char key[65];
	int l1, l2, decryp = 1;				/* decrypt */
	char buf1[128], buf2[128];

	DeDumpLog = stdout;
	
	l1 = strlen(strcpy(buf1, "Hello, World!\n"));
	l2 = strlen(strcpy(buf2, "I said: Hello, World!!!\n"));
	strcpy(key, "0123456789abcdef");

	de_setkeys(key, key);
	de_on(1);
	de_encrypt(1, buf1, l1);
	de_encrypt(1, buf2, l2);
	de_decrypt(1, buf1, l1);
	de_decrypt(1, buf2, l2);
	/* but what about encrypting in two hunks and decrypting in one */
	de_setkeys(key, key);
	de_encrypt(1, buf1, 3);
	de_encrypt(1, &buf1[3], l1 - 3);
	de_decrypt(1. buf1, l1);
	exit(0);
}
#endif
