/* NAME:
 *	hpw.c - authenticate user via hashed passwd
 *
 * DESCRIPTION:
 *	This is an authentication method that avoids transmitting
 *	passwords or other useful data across an in-secure network in
 *	the clear.  
 *	
 *	It uses UNIX hashed passwords and only the salt is transmitted
 *	in the clear and there is little or no fixed form data so
 *	there is little food for off-line dictionary attacks.   A
 *	private read-only passwd file is also supported. 
 *	
 *	Note that there is no reliance on any ``trusted'' third party.
 *	If the hosts at either end of the connection are secure, this
 *	authentication method should not make them much worse off.
 *
 *	There are stronger and more secure authentication options
 *	available.  This one simply requires little or no setup and
 *	avoids the biggest problem of telnet - transmitting clear text
 *	passwords. 
 *
 *	Protocol:
 *	
 *	Basically, the client says: 'AUTH' "user"
 *	
 *	The server looks up the hashed password for "user" and sends:
 *	'CHALLENGE' "version","salt","fs","echallenge","fs"
 *	where:
 *
 *	"version"
 *		indicates which version of HPW is being used.  This is
 *		for future use.
 *
 *	"salt"	is the salt used in the stored encrypted passwd.
 *	
 *	"fs"	is a random field separator (range is 0x1-0x1f).
 *
 *	"echallenge"
 *		contains "chtype","md","challenge","junk". "md" is an
 *		MD5 digest of the encrypted passwd ("hpw"), the actual
 *		"challenge" and some random "junk" to pad the message
 *		to 128 bytes.
 *
 *	The entire "echallenge" string is encrypted using "hpw" as
 *	the key.  Note that the "salt" which is transmitted in the
 *	clear is NOT used as part of the key.
 *
 *	The client prompts the user for their secret password and hashes
 *	it with "salt". It can then decrypt "echallenge".  The client
 *	calculates an MD5 digest of its own and compares it to that
 *	contained in "echallenge".  If they match, the client knows
 *	that the server has "hpw".
 *	
 *	The "chtype" is a random byte with only bit 0 significant.
 *	If bit 0 is clear it indicates that "challenge" is a simple
 *	random number (requiring response of number + 1).  In this
 *	case the client does not involve the user.
 *
 *	If "chtype" bit 0 is set then "challenge" is assumed to be from a One
 *	Time Password system and the client passes it on to the user
 *	and gathers the response.
 *
 *	Once the client as a "response" to send, it encrypts
 *	"fs","md","response" padded with random data with "hpw" (minus
 *	the "salt") and returns it to the server:
 *	'RESPONSE' "eresponse". 
 *
 *	The server decrypts "eresponse", checks that the contained
 *	MD5 digest matches its own and that "response" satisfies
 *	"challenge".  If so then the client has been authenticated.
 *
 *	Both sides can now set a session key using the MD5 digest to
 *	encrypt the decrypted "eresponse"  and using the result to set
 *	separate read and write session keys.  
 *
 * BUGS:
 *	On 4.4BSD systems, only processes with euid==0 can get hashed
 *	passwords from /etc/passwd (this is a feature).
 *	It annoys users having to answer multiple challenges.
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1995, Simon J. Gerraty.
 *      
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */

#ifndef lint
static char RCSid[] = "$Id: hpw.c,v 1.6 1997/05/05 03:11:50 sjg Exp $";
#endif
#ifdef	AUTH_HPW
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <stdio.h>
#include <pwd.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <arpa/telnet.h>
#include <pwd.h>
#include <stdio.h>

#ifdef	__STDC__
#include <stdlib.h>
#endif
#ifdef	NO_STRING_H
#include <strings.h>
#else
#include <string.h>
#endif

#include "encrypt.h"
#include "auth.h"
#include "misc.h"

#include "auth.h"
#include "des.h"
#include "md5.h"

#include "desenc.h"

#ifdef MAIN
# undef EXTERN
# define EXTERN
#endif
#ifndef EXTERN
# define EXTERN extern
#endif
EXTERN char *MyId, *Myname;
#ifdef AUTH_OLDHPW
EXTERN int UseFWTK;
#else
int UseFWTK = 0;
#endif

#define ARRSTR(a, s) (strncpy(a, s, sizeof (a) - 1), (a)[sizeof(a) - 1] = '\0', (a))

static u_char s_md[MD5_DIGEST_LENGTH];

static unsigned char str_data[2048] = { IAC, SB, TELOPT_AUTHENTICATION, 0,
			  		AUTHTYPE_HPW, };
static unsigned char str_name[1024] = { IAC, SB, TELOPT_AUTHENTICATION,
					TELQUAL_NAME, };
static u_char tmp_buf[1024];

#define	HPW_AUTH	0	/* Authentication data follows */
#define	HPW_REJECT	1	/* Rejected (reason might follow) */
#define HPW_ACCEPT	2	/* Accepted */
#define	HPW_CHALLENGE	3	/* Challenge for mutual auth. */
#define	HPW_ACK		4	/* Acknowledge */

#define HPW_VERSION	1			/* one byte */


static int hpw_version = 0;			/* assume the worst */

static struct passwd *Pw = NULL;

#define HPWSZ 13				/* size of hashed passwd */

struct ECH {
	char	e_fs;
	char	e_md[MD5_DIGEST_LENGTH];
	char	e_data[1];
};

#define HPW_MSGSZ	(128 - sizeof(struct ECH))
	
static void
hpw_setkeys(seed, nbytes, server)
	char *seed;
	int nbytes;
	int server;
{
	char tmp[128];

	/*
	 * note that we do not use the 1st 2 chars as
	 * these are the salt that was transmitted in the clear
	 */
	ARRSTR(tmp, seed);

	if (nbytes < 0)
		nbytes = strlen(seed);
	if (nbytes > sizeof(tmp))
		nbytes = sizeof(tmp);

	de_buf_key(s_md, tmp, tmp, MD5_DIGEST_LENGTH, DE_ENCRYPT);
	if (server)
		de_setkeys(tmp, &tmp[nbytes / 2]);
	else
		de_setkeys(&tmp[nbytes / 2], tmp);
}


static unsigned long RSeed;

static char *
null_challenge(user, typep, err)
	char *user;
	int *typep;
	char **err;
{
	static char buf[32];
	time_t now = time((time_t *) 0);
	
	if (typep)
		*typep = 0;			/* numeric challenge */
	srand(now);
	RSeed = rand() * getpid();
	(void) sprintf(buf, "%lu", RSeed);

	return buf;
}

static int
null_verify(user, response, err)
	char *user, *response, **err;
{
	unsigned long ul;

	ul = atol(response);

	return (ul == (RSeed + 1));
}

/* get response(s) from user */
char *
hpw_prompt(msg)
	char *msg;
{
	extern char * telnet_gets();
	extern char * crypt();
	extern char * getpass();
	char    pwprompt[128], tmp[1024], buf[1024], hpw[16], salt[16];
	char    *p, *emsg, *challenge, *response;
	int	fs, i, passok, n, type;
	unsigned long ul;
	struct ECH *ecp;
	MD5_CTX c;
	u_char md[MD5_DIGEST_LENGTH];
	
	if (msg == NULL)
		return NULL;

	hpw_version = *msg++;			/* for future */
	
	ARRSTR(tmp, msg);

	hpw[0] = fs = tmp[2];			/* 3rd char */
	hpw[1] = '\0';
	
	if ((p = strtok(tmp, hpw)) == NULL ||
	    (emsg = strtok(NULL, hpw)) == NULL)
		return NULL;

	ARRSTR(salt, tmp);
	response = NULL;
	passok = 0;
	n = x2b(tmp_buf, emsg, -1);
	challenge = tmp_buf;
	
	/*
	 * Get user's secret password
	 * We include MyId in prompt so user can tell its us...
	 */
	
	(void) sprintf(pwprompt, "%s: Password: ", MyId);
	
	for (i = 0; i < 2 && passok == 0; i++) {

		p = crypt(telnet_gets(pwprompt, buf, sizeof(buf), 0), salt);
		ARRSTR(hpw, p);

		memcpy(tmp, challenge, n);
		/* don't use salt in key */
		de_buf_key(&hpw[2], tmp, tmp, n, DE_DECRYPT);

		ecp = (struct ECH *) tmp;
		
		type = ecp->e_fs & 1;

		memcpy(s_md, ecp->e_md, MD5_DIGEST_LENGTH); /* save it */
		/* calulate our md5 digest... */
		MD5_Init(&c);
		MD5_Update(&c, hpw, strlen(hpw));
		MD5_Update(&c, ecp->e_data, HPW_MSGSZ);
		MD5_Final(&(md[0]),&c);

		if (memcmp(s_md, md, MD5_DIGEST_LENGTH) == 0) {
			passok = 1;
			challenge = ecp->e_data;
			if (p = strchr(challenge, fs))
				*p = '\0';
			break;
		}
		sleep(2);			/* annoy them */
	}
	if (passok) {
		switch (type) {
		case 0:
			ul = atoi(challenge);
			(void) sprintf(tmp, "%lu", ul + 1);
			response = tmp;
			break;
		case 1:
			/* need to prompt user and get response */
			if ((p = strstr(challenge, "esponse")) != NULL) {
				p = (strchr(p, ':') != NULL) ? "" : ": ";
				(void) sprintf(buf, "%s: %s%s",
					       MyId, challenge, p);
			} else {
				(void) sprintf(buf, "%s: challenge <%s> response: ",
					       MyId, challenge);
			}
			response = telnet_gets(buf, tmp, sizeof(tmp), 1);
			break;
			
		}
		if (response != NULL) {
			struct ECH *ecp = (struct ECH *)buf;
			ecp->e_fs = fs;
			memcpy(ecp->e_md, s_md, MD5_DIGEST_LENGTH);
			(void) sprintf(ecp->e_data, "%.256s%c",
				       response, fs);

			n = sizeof (struct ECH) + strlen(response);
			memcpy(tmp_buf, buf, n);
			if (n < HPW_MSGSZ) {
				int i;

				/*
				 * pad out to HPW_MSGSZ with random data
				 */
				for (i = n; i < HPW_MSGSZ + 1;
				     i += MD5_DIGEST_LENGTH) {
					MD5_Update(&c, md, MD5_DIGEST_LENGTH);
					MD5_Final(&(md[0]),&c);
					memcpy(&tmp_buf[i], md, MD5_DIGEST_LENGTH);
				}
				n = HPW_MSGSZ;
			}
			de_buf_key(&hpw[2], tmp_buf, tmp_buf, n, DE_ENCRYPT);
			
			hpw_setkeys(buf, n, 0);	/* unencrytped response */
	
			return b2x(buf, tmp_buf, n);
		}
	}
	de_setstatus(0);			/* not happy */
	return NULL;
}


/* verify response */
int
hpw_verify(user, verify, response, err)
	char *user, *response;
	int (*verify)();
	char **err;
{
	int fs, n, sts = 0;
	char *p, *resp, tmp[512];

	if (Pw == NULL || response == NULL)
		return 0;
	n = x2b(tmp_buf, response, -1);
	resp = tmp_buf;
	/* don't use salt */
	de_buf_key(&(Pw->pw_passwd)[2], resp, resp, n, DE_DECRYPT);

	memcpy(tmp, resp, n);			/* need to remember this */

	
	fs = *resp++;
	if (memcmp(s_md, resp, MD5_DIGEST_LENGTH) == 0) {
		resp += MD5_DIGEST_LENGTH;
		if (p = strchr(resp, fs))
			*p = '\0';
		
		if (verify == NULL)
			verify = null_verify;
		sts = (verify(user, resp, err) > 0);
	}
	if (sts > 0) {
		/*
		 * truncate it at fs
		 */
		if (p = strchr(&tmp[MD5_DIGEST_LENGTH + 1], fs))
			*++p = '\0';
		hpw_setkeys(tmp, n, 1);
	} else {
		de_setstatus(0);		/* not happy */
	}
	return sts;
}

char *
hpw_challenge(user, type, challenge)
	char *user;
	int type;
	char *challenge;
{
	extern struct passwd *telnet_getpw();
	static char buf[1024];
        char tmp[1024];
	struct ECH *ecp;
	MD5_CTX c;
	u_char md[MD5_DIGEST_LENGTH];
	int j, n;
	int fs;
	
	if (user != NULL &&
	    (Pw = telnet_getpw(user)) != NULL &&
	    strlen(Pw->pw_passwd) >= HPWSZ) {


		if (challenge == NULL) {
			challenge = null_challenge(user, &type, (char **)0);
		}
		(void) sprintf(tmp, "%s%ld", MyId, time((time_t *)0));
		MD5_Init(&c);
		/* first some random junk */
		MD5_Update(&c, tmp, strlen(tmp));
		MD5_Final(&(md[0]),&c);

		for (n = 0; n < 256; n += MD5_DIGEST_LENGTH) {
			MD5_Update(&c, md, MD5_DIGEST_LENGTH);
			MD5_Final(&(md[0]),&c);
			memcpy(&buf[n], md, MD5_DIGEST_LENGTH);
		}

		ecp = (struct ECH *) tmp;

		fs = (md[1] | 0x2) & 0x1f;	/* random field sep? */
		
		if (type) {
			fs |= 1;
		} else {
			fs &= ~1;
		}
		ecp->e_fs = fs;
		(void) sprintf(ecp->e_data, "%.*s%c", HPW_MSGSZ, challenge, fs);
		n = strlen(challenge);
		(void) memcpy(&ecp->e_data[n+1], buf, HPW_MSGSZ); /* junk */

		MD5_Init(&c);
		MD5_Update(&c, Pw->pw_passwd, strlen(Pw->pw_passwd));
		MD5_Update(&c, ecp->e_data, HPW_MSGSZ);
		MD5_Final(&(md[0]),&c);

		memcpy(ecp->e_md, md, MD5_DIGEST_LENGTH);
		memcpy(s_md, md, MD5_DIGEST_LENGTH); /* save it */
		n = HPW_MSGSZ + sizeof (struct ECH);
		/* don't use salt */
		de_buf_key(&(Pw->pw_passwd)[2], tmp, tmp, n, DE_ENCRYPT);
		
		(void) sprintf(buf, "%c%.2s%c%s%c",
			       HPW_VERSION,
			       Pw->pw_passwd, fs,
			       b2x(tmp_buf, tmp, n), fs);
#ifdef MAIN
		printf("n = %d, strlen(buf) = %d\n", n, strlen(buf));
#endif
		return buf;
        }
	return NULL;
}


#ifdef MAIN
char *
telnet_gets(pwprompt, buf, nbytes, f)
	char *pwprompt, *buf;
	int nbytes, f;
{
	return strncpy(buf, getpass(pwprompt), nbytes);
}

int
vrfy(u, r)
	char *u, *r;
{
	printf("User(%s) says: '%s'\n", u, r);
	return 1;
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int optind, getopt();
	extern char *optarg, *_path_authpasswd;
	static char myid[128];
	extern FILE *DeDumpLog;
	register char *rcp, *p;
	char *user = "root";
	int i, type = 0;
	char *err = NULL;
	
	DeDumpLog = stdout;

	if ((rcp = strrchr(argv[0], '/')))
		rcp++;
	else
		rcp = argv[0];
	
	sprintf(myid, "%s[%d]", rcp, getpid());
	MyId = myid;

	printf("My id is: %s\n", MyId);

	while ((i = getopt(argc, argv, "t:p:u:")) != EOF) {
		switch (i) {
		case 't':
			type = atoi(optarg);
			break;
		case 'p':
			_path_authpasswd = optarg;
			break;
		case 'u':
			user = optarg;
			break;
		}
	}

	if (rcp = hpw_challenge(user, type,
				   (type) ? "What is your quest" :
				   NULL)) {

		/* now get response */
		for (i = 0; i < 5; i++) {
			/* simulate network exchange... */
			if (p = hpw_prompt(rcp))
				p = strdup(p);
			if (hpw_verify(user,
					  (type) ? vrfy : NULL, p, &err) > 0) {
				printf("Response OK\n");
				free(p);
				break;
			}
			free(p);
		}
	}

	return 0;
}

#else /* MAIN */

static int
Data(ap, type, d, c)
	Authenticator *ap;
	int type;
	void *d;
	int c;
{
        unsigned char *p = str_data + 4;
	unsigned char *cd = (unsigned char *)d;

	if (c == -1)
		c = strlen((char *)cd);

        if (0) {
                fprintf(stderr,"%s:%d: [%d] (%d)",
                        str_data[3] == TELQUAL_IS ? ">>>IS" : ">>>REPLY",
                        str_data[3],
                        type, c);
                fprintd(stderr,d, c);
                fprintf(stderr,"\r\n");
        }
	*p++ = ap->type;
	*p++ = ap->way;
	*p++ = type;
        while (c-- > 0) {
                if ((*p++ = *cd++) == IAC)
                        *p++ = IAC;
	}
        *p++ = IAC;
        *p++ = SE;
	if (str_data[3] == TELQUAL_IS)
		printsub('>', &str_data[2], p - (&str_data[2]));
        return(net_write(str_data, p - str_data));
}

static char hpw_buf[1024];

int
hpw_init(ap, server)
	Authenticator *ap;
	int server;
{
	de_off(-1);
	
	SetMyId(server);
	
	if (server) {
		str_data[3] = TELQUAL_REPLY;
	} else {
		str_data[3] = TELQUAL_IS;
	}
	return(1);
}

int
hpw_send(ap)
	Authenticator *ap;
{
	static char version[4] = { HPW_VERSION, 0, 0, 0, };
	char *cp;
	
	if (!UserNameRequested) {
		if ((cp = getenv("USER")) == NULL)
			return 0;
		
		(void) printf("%s: no -l user specified, assuming %s\n",
			      MyId, cp);
		auth_encrypt_user(cp);
	}
	printf("%s: Trying HPW '%s' ...\n", MyId, UserNameRequested);
	if (!auth_sendname(UserNameRequested, strlen(UserNameRequested))) {
		return(0);
	}

	if (!Data(ap, HPW_ACK, (void *)version, 2)) {
		return(0);
	}

	return(1);
}

		
/* server receives IS */
void
hpw_is(ap, data, cnt)
	Authenticator *ap;
	unsigned char *data;
	int cnt;
{
	char *cp;
	char *err = NULL;
	int (*verify)() = NULL;
	
	if (cnt-- < 1)
		return;

	switch (*data++) {
	case HPW_AUTH:
		bcopy((void *)data, (void *)hpw_buf, cnt);
		hpw_buf[cnt] = '\0';
#ifdef USE_FWTK
		if (UseFWTK) {
			extern int fwtk_verify __P((char *user, char *response, char **err));
			verify = fwtk_verify;
		}
#endif
		if (hpw_verify(UserNameRequested, verify, hpw_buf, &err) <= 0) {
			if (err == NULL)
				err = "Auth failed";
			Data(ap, HPW_REJECT, (void *)err, -1);
			auth_finished(ap, AUTH_REJECT);
			return;
		}
		Data(ap, HPW_ACCEPT, (void *)0, 0);
		auth_finished(ap, AUTH_USER);
		break;

	case HPW_CHALLENGE:
		break;


	case HPW_ACK:
		/* should have user name by now... */
		if (UserNameRequested) {
			int type = 0;
			char *challenge = NULL;

#ifdef USE_FWTK
			if (UseFWTK) {
				extern char *fwtk_challenge __P((char *user, char *prog, char **err));
				char *err;
				
				type = 1;
				alarm(0);	/* no timeout */
				challenge = fwtk_challenge(UserNameRequested, Myname, &err);
			}
#endif
			hpw_version = *data;
			cp = hpw_challenge(UserNameRequested, type, challenge);
			if (cp == NULL) {
				/* don't clue them in, give them a fake */
				cp = "42*0*a9,7d,9f,9d,50,75,1a,*";
			}
			Data(ap, HPW_CHALLENGE, (void *)cp, strlen(cp));
		}
		break;
		
	default:
		Data(ap, HPW_REJECT, 0, 0);
		break;
	}
}


/* this what the client sees */
void
hpw_reply(ap, data, cnt)
	Authenticator *ap;
	unsigned char *data;
	int cnt;
{
	char *cp;
	
	if (cnt-- < 1)
		return;
	switch (*data++) {
	case HPW_REJECT:
		if (cnt > 0) {
			printf("%s: HPW refuses authentication because %.*s\r\n",
				MyId, cnt, data);
		} else
			printf("%s: HPW refuses authentication\r\n", MyId);
		auth_send_retry();
		return;
	case HPW_ACCEPT:
		printf("%s: HPW accepts you\n", MyId);
		auth_finished(ap, AUTH_USER);
		return;
	case HPW_CHALLENGE:
		hpw_version = *data;
		bcopy((void *)data, (void *)hpw_buf, cnt);
		hpw_buf[cnt] = '\0';

		if ((cp = hpw_prompt(hpw_buf)) == NULL) {
			cp = "sorry about that";
		}
		Data(ap, HPW_AUTH, (void *)cp, strlen(cp));
		break;

	default:
		return;
	}
}

int
hpw_status(ap, name, level)
	Authenticator *ap;
	char *name;
	int level;
{
	int st;
	
	if (level < AUTH_USER)
		return(level);
	
	st = de_status(NULL);
	if (UserNameRequested && (st & DE_KEYSET)) {
		strcpy(name, UserNameRequested);
		return(AUTH_VALID);
	}
	return(AUTH_USER);
}

#define	BUMP(buf, len)		while (*(buf)) {++(buf), --(len);}
#define	ADDC(buf, len, c)	if ((len) > 0) {*(buf)++ = (c); --(len);}

void
hpw_printsub(data, cnt, buf, buflen)
	unsigned char *data, *buf;
	int cnt, buflen;
{
	char lbuf[32];
	register int i;

	buf[buflen-1] = '\0';		/* make sure its NULL terminated */
	buflen -= 1;

	switch(data[3]) {
	case HPW_REJECT:	/* Rejected (reason might follow) */
		strncpy((char *)buf, " REJECT ", buflen);
		goto common;

	case HPW_ACCEPT:	/* Accepted (name might follow) */
		strncpy((char *)buf, " ACCEPT ", buflen);
	common:
		BUMP(buf, buflen);
		if (cnt <= 4)
			break;
		ADDC(buf, buflen, '"');
		for (i = 4; i < cnt; i++)
			ADDC(buf, buflen, data[i]);
		ADDC(buf, buflen, '"');
		ADDC(buf, buflen, '\0');
		break;

	case HPW_AUTH:		/* Authentication data follows */
		strncpy((char *)buf, " AUTH", buflen);
		goto common2;

	case HPW_CHALLENGE:
		strncpy((char *)buf, " CHALLENGE", buflen);
		goto common2;

	case HPW_ACK:
		strncpy((char *)buf, " ACK", buflen);
		goto common2;

	default:
		sprintf(lbuf, " %d (unknown)", data[3]);
		strncpy((char *)buf, lbuf, buflen);
	common2:
		BUMP(buf, buflen);
		for (i = 4; i < cnt; i++) {
			sprintf(lbuf, " %d", data[i]);
			strncpy((char *)buf, lbuf, buflen);
			BUMP(buf, buflen);
		}
		break;
	}
}
#endif
#endif
